/*
    Fingerprint selection algorithms for local text reuse detection
    Copyright (C) 2020 Gints Jekabsons, gints.jekabsons@rtu.lv
    Available at: http://www.cs.rtu.lv/jekabsons/nlp.html
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
	
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
	
    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

package lv.rtu.fingerprinting;

public class Configuration {
	public int verbose = 0; // 0 = quiet; 1 = reporting false positives; 2 = reporting words and hashes
	
	public boolean removeStopWords;
	public int minWordLength;
	public boolean doStemming;
	public boolean nGramsConsistOfWords;
	public int gramSize;
	
	public static enum Algorithm {FULL, EVERY_PTH, MODP, WINNOWING, FBW, MFBW, HAILSTORM};
	public Algorithm algorithm;
	
	public int parameter;
	public int k;
	public int nTop;
	
	/**
	 * Configuration for fingerprinting.
	 * @param removeStopWords - whether to remove stop words.
	 * @param minWordLength - what is the minimum allowed word (token) length in a document.
	 * @param doStemming - whether to do stemming.
	 * @param nGramsConsistOfWords - whether n-grams should be made of words instead of characters.
	 * @param gramSize - what should be the n-gram size (this is the n parameter).
	 * @param algorithm - which fingerprint selection algorithm should be used.
	 * @param parameter - parameter value for the fingerprint selection algorithm (has no effect for FULL and HAILSTORM algorithms).
	 * @param k - how many consecutive n-grams to search for in a single query.
	 * @param nTop - how many top documents should Lucene retrieve for a query result.
	 */
	public Configuration(
			boolean removeStopWords,
			int minWordLength,
			boolean doStemming,
			boolean nGramsConsistOfWords,
			int gramSize,
			Algorithm algorithm,
			int parameter,
			int k,
			int nTop) {
		this.removeStopWords = removeStopWords;
		this.minWordLength = minWordLength;
		this.doStemming = doStemming;
		this.nGramsConsistOfWords = nGramsConsistOfWords;
		this.gramSize = gramSize;
		this.algorithm = algorithm;
		this.parameter = parameter;
		this.k = k;
		this.nTop = nTop;
	}
	
	public boolean needFreq() {
		return (algorithm == Algorithm.FBW) || (algorithm == Algorithm.MFBW);
	}
	
	@Override
	public Configuration clone() {
		return new Configuration(removeStopWords, minWordLength, doStemming, nGramsConsistOfWords, gramSize, algorithm, parameter, k, nTop);
	}
	
	@Override
	public String toString() {
		return removeStopWords + "\t" + minWordLength + "\t" + doStemming + "\t" + nGramsConsistOfWords + "\t" +
				gramSize + "\t" + algorithm + "\t" + parameter + "\t" + k + "\t" + nTop;
	}
	
	public String getColumnNames() {
		return "Remove stop words\tMin word length\tStemming\tWord n-grams\tGram size\tAlgorithm\tParameter\tk\tnTop";
	}
	
	public String getResultsFileName() {
		return "Result - " + algorithm.toString().toLowerCase() + ", remStop=" + (removeStopWords ? 't' : 'f') +
				", minLen=" + minWordLength + ", stem=" + (doStemming ? 't' : 'f');
	}
}
