/*
    Fingerprint selection algorithms for local text reuse detection in a two-stage system
    Copyright (C) 2021 Gints Jekabsons, gints.jekabsons@rtu.lv
    Available at: http://www.cs.rtu.lv/jekabsons/nlp.html
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
	
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
	
    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

package lv.rtu.fingerprinting;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.core.WhitespaceAnalyzer;
import org.apache.lucene.document.DocumentStoredFieldVisitor;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.IndexWriterConfig;
import org.apache.lucene.index.Term;
import org.apache.lucene.search.BooleanClause;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.MMapDirectory;
import org.apache.lucene.util.BytesRef;

/**
 * <p>
 * <h1>Fingerprint selection algorithms for local text reuse detection in a two-stage system</h1>
 * </p>
 * <p>
 * Developed by Gints Jekabsons.
 * <br>
 * Available at: <a href="http://www.cs.rtu.lv/jekabsons/nlp.html">http://www.cs.rtu.lv/jekabsons/nlp.html</a>.
 * </p>
 * <p>
 * This software was used for the experiments in the paper
 * "X"
 * (available: <a href="http://www.cs.rtu.lv/jekabsons/Files/Jek_ACSS2021.pdf">http://www.cs.rtu.lv/jekabsons/Files/Jek_ACSS2021.pdf</a> and <a href="https://sciendo.com/article/10.2478/acss-2021-0022">https://sciendo.com/article/10.2478/acss-2021-0022</a>).
 * <br>
 * <b>If you are using this software, please give a reference the paper.</b>
 * <br>
 * And it would be nice to give a link to the website as well: <a href="http://www.cs.rtu.lv/jekabsons/nlp.html">http://www.cs.rtu.lv/jekabsons/nlp.html</a>.
 * </p>
 * <p>
 * This software is developed for evaluating the effectiveness of fingerprint selection algorithms for a two-stage (source retrieval + aligning) local text reuse detection.
 * It implements the following fingerprint selection algorithms (for details on the algorithms see <a href="https://content.sciendo.com/view/journals/acss/25/1/article-p11.xml">https://content.sciendo.com/view/journals/acss/25/1/article-p11.xml</a> and <a href="https://sciendo.com/article/10.2478/acss-2021-0022">https://sciendo.com/article/10.2478/acss-2021-0022</a>):
 * <li>Full fingerprinting;</li>
 * <li>Every p-th;</li>
 * <li>0 mod p;</li>
 * <li>Winnowing;</li>
 * <li>Hailstorm;</li>
 * <li>Frequency-Biased Winnowing (FBW);</li>
 * <li>Modified Frequency-Biased Winnowing (MFBW).</li>
 * </p>
 * <br>
 * <p>
 * Indexing of the fingerprints is implemented using the <a href="https://lucene.apache.org/">Apache Lucene library</a>.
 * </p>
 * <br>
 * <p>
 * To use the software, you need a dataset containing collection documents (directory 'src') and query documents (directory 'susp').
 * You can set the directories for the files using the static variables of the FingerprintIndexing class.
 * <br>
 * You also need to create a directory named 'correct' that contains xml files for all correct text reuse document pairs.
 * The xml files should contain data about overlapping text spans between the documents. The syntax for the xml files is the one defined at the PAN shared tasks.
 * <br>
 * You will also need an aligner (e.g., <a href="https://www.gelbukh.com/plagiarism-detection/PAN-2015/index.html">the one developed by Sanchez-Perez</a>) and an evaluator (e.g., the one developed for the PAN 2009-2014 shared task; see 'evaluate.ipynb').
 * </p>
 * <br>
 * <p>
 * Java dependencies: <a href="https://opennlp.apache.org/">Apache OpenNLP</a> (used with version 1.9.0) and <a href="https://lucene.apache.org/">Apache Lucene</a> (used with version 8.8.0).
 * </p>
 * <br>
 * @version 1.0 (December 30, 2021)
 * @author Gints Jekabsons, gints.jekabsons@rtu.lv
 */

public class FingerprintIndexing {
	private static final String DATA_PATH = "G:/_/Data/";					// Main directory containing data subdirectories
	private static final String CORRECT_PATH = DATA_PATH + "correct/";		// Directory for correct information about pairs
	private static final String QUERY_PATH = DATA_PATH + "susp/";			// Directory for query documents to be checked against the collection documents
	private static final String COLLECTION_PATH = DATA_PATH + "src/";		// Directory for collection documents
	
	private static final String INDICES_PATH = "G:/_/Idx/";					// Main directory where to store indices
	private static final String RESULTS_PATH = "G:/_/Results/";				// Main directory for storing results
	
	private static final String FIELD_FILENAME = "filename";
	private static final String FIELD_FINGERPRINT = "fingerprint";
	
	public static void main(String[] args) throws Exception {
		(new FingerprintIndexing()).start();
	}
	
	public void start() throws Exception {
		Configuration config = new Configuration(true, 3, true, true, 1, Configuration.Algorithm.FULL, 0, 1, 1);
		
		System.out.println("Getting file lists...");
		
		List<File> queryFiles = getDirectoryFileList(QUERY_PATH);
		List<File> collectionFiles = getDirectoryFileList(COLLECTION_PATH);
		
		// Here we prepare a map from query documents to their correct sources in the collection
		Map<String, HashSet<String>> correctPairs = new HashMap<String, HashSet<String>>();
		List<File> correctFiles = getDirectoryFileList(CORRECT_PATH);
		for (File pairFile : correctFiles) {
			String[] filenames = pairFile.getName().split("-");
			String queryFilename, collectionFilename;
			if (filenames.length == 2) {
				queryFilename = filenames[0] + ".txt";
				collectionFilename = filenames[1].substring(0, filenames[1].length() - 4) + ".txt";
			} else
				if (filenames.length == 4) {
					queryFilename = filenames[0] + "-" + filenames[1] + ".txt";
					collectionFilename = filenames[2] + "-" + filenames[3].substring(0, filenames[3].length() - 4) + ".txt";
				} else
					continue;
			
			String text = new String(Files.readAllBytes(pairFile.toPath()), "UTF-8");
			if (!text.contains("this_offset"))
				continue;
			
			HashSet<String> set = correctPairs.get(queryFilename);
			if (set == null) {
				set = new HashSet<String>();
				set.add(collectionFilename);
				correctPairs.put(queryFilename, set);
			} else
				set.add(collectionFilename);
		}
		
		System.out.println("Preloading documents...");
		
		// If you have too many documents to preload all of them in the memory,
		// just set preloadedQueryDocs and preloadedCollectionDocs to null and comment-out both for-loops.
		
		Map<File, Document> preloadedQueryDocs = new HashMap<File, Document>();
		for (File file : queryFiles)
			preloadedQueryDocs.put(file, new Document(file, config));
		
		Map<File, Document> preloadedCollectionDocs = new HashMap<File, Document>();
		for (File file : collectionFiles)
			preloadedCollectionDocs.put(file, new Document(file, config));
		
		System.out.println("Running experiments...");
		
		OutputStreamWriter outputWriter = new OutputStreamWriter(new FileOutputStream(RESULTS_PATH + config.getResultsFileName() + ".txt"), StandardCharsets.UTF_8);
		outputWriter.write(config.getColumnNames());
		outputWriter.write("\tRetrieved\tPrecision\tRecall\tF1\tAll tokens\tAll n-grams\tFingerprints\tUnique fingerprints\tIndex size\r\n");
		
		for (config.gramSize = 1; config.gramSize <= 16; config.gramSize++) { //XXX set in accordance to your needs
			
			// If n-gram collection frequencies are required, here we compute them in the collection documents by doing "Full fingerprinting"
			HashMap<Integer, Integer> hashFreq = null;
			if (config.needFreq()) {
				Configuration configBase = config.clone();
				configBase.algorithm = Configuration.Algorithm.FULL;
				
				hashFreq = new HashMap<Integer, Integer>();
				if (preloadedCollectionDocs != null) {
					for (Document doc : preloadedCollectionDocs.values())
						doc.createFingerprints(configBase, hashFreq, true);
				} else {
					for (File file : collectionFiles)
						(new Document(file, config)).createFingerprints(configBase, hashFreq, true);
				}
			}
			
			int[] params;
			if ((config.algorithm == Configuration.Algorithm.FULL) ||
				(config.algorithm == Configuration.Algorithm.HAILSTORM)) {
				// We will not loop because these two algorithms do not have the parameter
				params = new int[] {0};
			} else {
				if ((config.algorithm == Configuration.Algorithm.WINNOWING) ||
					(config.algorithm == Configuration.Algorithm.FBW) ||
					(config.algorithm == Configuration.Algorithm.MFBW))
					params = new int[] {3, 5, 9, 19, 39}; //XXX set in accordance to your needs
				else
					params = new int[] {2, 3, 5, 10, 20}; //XXX set in accordance to your needs
			}
			
			for (int param : params) {
				config.parameter = param;
				
				// Here we compute some statistics and, if the documents are preloaded, they now will have their fingerprints ready
				int numAllTokens = 0;
				int numAllNGrams = 0;
				int numFingerprints = 0;
				int numUniqueFingerprints = 0;
				if (preloadedCollectionDocs != null) {
					for (Document doc : preloadedCollectionDocs.values()) {
						doc.createFingerprints(config, hashFreq, false);
						numAllTokens += doc.getNumAllTokens();
						numAllNGrams += doc.getNumAllNGrams();
						numFingerprints += doc.getNumFingerprints();
						numUniqueFingerprints += doc.getNumUniqueFingerprints();
					}
				} else {
					for (File file : collectionFiles) {
						Document doc = new Document(file, config);
						doc.createFingerprints(config, hashFreq, false);
						numAllTokens += doc.getNumAllTokens();
						numAllNGrams += doc.getNumAllNGrams();
						numFingerprints += doc.getNumFingerprints();
						numUniqueFingerprints += doc.getNumUniqueFingerprints();
					}
				}
				
				if (preloadedQueryDocs != null) {
					for (Document doc : preloadedQueryDocs.values())
						doc.createFingerprints(config, hashFreq, false);
				}
				
				String dirIndex = config.algorithm + "_" + config.gramSize + "_" + config.parameter + "/";
				
				File directory = new File(INDICES_PATH + dirIndex);
				if (directory.exists()) {
					System.out.println("Directory \"" + directory + "\" already exists.");
					outputWriter.close();
					return;
				}
				if (!directory.mkdir()) {
					outputWriter.close();
					return;
				}
				
				index(INDICES_PATH + dirIndex, collectionFiles, preloadedCollectionDocs, hashFreq, config);
				
				// Compute index size (without .pos file as pos info is actually not required and should be disabled)
				List<File> indexFiles = getDirectoryFileList(INDICES_PATH + dirIndex);
				long indexSize = 0;
				for (File file : indexFiles)
					if (!file.getName().endsWith(".pos"))
						indexSize += file.length();
				
				int firstK = 1; //XXX set in accordance to your needs
				int lastK = 1; //XXX set in accordance to your needs
				
				for (config.k = firstK; config.k <= lastK; config.k++) {
					
					System.out.println("gramSize=" + config.gramSize + ", parameter=" + config.parameter + ", k=" + config.k);
					
					String dirResult = config.algorithm + "_" + config.gramSize + "_" + config.parameter + "_" + config.k + "/";
					
					directory = new File(RESULTS_PATH + dirResult);
					if (directory.exists()) {
						System.out.println("Directory \"" + directory + "\" already exists.");
						outputWriter.close();
						return;
					}
					if (!directory.mkdir()) {
						outputWriter.close();
						return;
					}
					
					directory = new File(RESULTS_PATH + dirResult + "correct/");
					if (!directory.mkdir()) {
						outputWriter.close();
						return;
					}
					directory = new File(RESULTS_PATH + dirResult + "out/");
					if (!directory.mkdir()) {
						outputWriter.close();
						return;
					}
					
					String results = search(INDICES_PATH + dirIndex, RESULTS_PATH + dirResult, CORRECT_PATH, queryFiles, preloadedQueryDocs, correctPairs, hashFreq, config);
					
					outputWriter.write(config.toString());
					outputWriter.write("\t");
					outputWriter.write(results);
					outputWriter.write("\t" + numAllTokens / collectionFiles.size() + "\t" + numAllNGrams / collectionFiles.size() + "\t" + numFingerprints / collectionFiles.size() + "\t" + numUniqueFingerprints / collectionFiles.size() + "\t" + indexSize);
					outputWriter.write("\r\n");
					outputWriter.flush();
					
				}
			}
		}
		
		outputWriter.close();
		
		System.out.println("Done");
	}
	
	private void index(
			String indexPath,
			List<File> collectionDocs,
			Map<File, Document> preloadedCollectionDocs,
			HashMap<Integer, Integer> hashFreq,
			Configuration config
			) throws InterruptedException, IOException {
		
		Directory indexDirectory = MMapDirectory.open(Paths.get(indexPath));
		
		Analyzer analyzer = new WhitespaceAnalyzer();
		IndexWriterConfig iwc = new IndexWriterConfig(analyzer);
		iwc.setRAMBufferSizeMB(512);
		IndexWriter writer = new IndexWriter(indexDirectory, iwc);
		
		final AtomicInteger idx = new AtomicInteger(0);
		
		final int threadCount = 6;
		ExecutorService executor = Executors.newFixedThreadPool(threadCount);
		ExecutorCompletionService<Integer> completionService = new ExecutorCompletionService<Integer>(executor);
		for (int i = 0; i < threadCount; i++) {
			Callable<Integer> task = () -> {
				indexFiles(writer, collectionDocs, preloadedCollectionDocs, idx, hashFreq, config);
				return null;
			};
			completionService.submit(task);
		}
		for (int i = 0; i < threadCount; i++) {
			try {
				completionService.take().get();
			} catch (Exception e) {
				e.printStackTrace();
				executor.shutdownNow();
				break;
			}
		}
		executor.shutdown();
		executor.awaitTermination(1, TimeUnit.DAYS);
		
		writer.commit();
		writer.forceMerge(1);
		
		writer.close();
		indexDirectory.close();
	}
	
	private int indexFiles(
			IndexWriter writer,
			List<File> collectionFiles,
			Map<File, Document> preloadedCollectionDocs,
			AtomicInteger idx,
			HashMap<Integer, Integer> hashFreq,
			Configuration config
			) throws Exception {
		
		org.apache.lucene.document.Document luceneDocument = new org.apache.lucene.document.Document();
		int numIndexedFiles = 0;
		while (true) {
			if (Thread.currentThread().isInterrupted())
				return numIndexedFiles;
			
			int i = idx.getAndIncrement();
			if (i >= collectionFiles.size())
				return numIndexedFiles;
			
			File collectionFile = collectionFiles.get(i);
			String filename = collectionFile.getName();
			
			Document collectionDoc = null;
			if (preloadedCollectionDocs != null)
				collectionDoc = preloadedCollectionDocs.get(collectionFile);
			if (collectionDoc == null) {
				collectionDoc = new Document(collectionFile, config);
				collectionDoc.createFingerprints(config, hashFreq, false);
			}
			
			luceneDocument.clear();
			StoredField filenameField = new StoredField(FIELD_FILENAME, filename);
			luceneDocument.add(filenameField);
			TextField fingerprintField = new TextField(FIELD_FINGERPRINT, collectionDoc.getFingerprintsAsString(), Field.Store.NO);
			luceneDocument.add(fingerprintField);
			
			writer.addDocument(luceneDocument);
		}
	}
	
	private String search(
			String indexPath,
			String resultsPath,
			String correctPath,
			List<File> queryFiles,
			Map<File, Document> preloadedQueryDocs,
			Map<String, HashSet<String>> correctPairs,
			HashMap<Integer, Integer> hashFreq,
			Configuration config
			) throws InterruptedException, IOException {
		
		Directory indexDirectory = MMapDirectory.open(Paths.get(indexPath));
		IndexReader reader = DirectoryReader.open(indexDirectory);
		IndexSearcher searcher = new IndexSearcher(reader);
		
		OutputStreamWriter owPairsCorrect = new OutputStreamWriter(new FileOutputStream(resultsPath + "correct/pairs"), StandardCharsets.UTF_8);
		
		int totalCorrectPairs = 0;
		for (Entry<String, HashSet<String>> entry : correctPairs.entrySet()) {
			totalCorrectPairs += entry.getValue().size();
			
			String queryFilename = entry.getKey();
			for (String collectionFilename : entry.getValue()) {
				owPairsCorrect.write(queryFilename);
				owPairsCorrect.write(" ");
				owPairsCorrect.write(collectionFilename);
				owPairsCorrect.write("\r\n");
				
				String xmlFilename = queryFilename.substring(0, queryFilename.length() - 4) + "-" + collectionFilename.substring(0, collectionFilename.length() - 4) + ".xml";
				File xmlFrom = new File(correctPath + xmlFilename);
				String text = new String(Files.readAllBytes(xmlFrom.toPath()), "UTF-8");
				OutputStreamWriter owTo = new OutputStreamWriter(new FileOutputStream(resultsPath + "correct/" + xmlFilename), StandardCharsets.UTF_8);
				owTo.write(text);
				owTo.close();
			}
		}
		
		OutputStreamWriter owPairsToAlign = new OutputStreamWriter(new FileOutputStream(resultsPath + "pairs_align"), StandardCharsets.UTF_8);
		
		int totalRetrievedFiles = 0;
		int tp = 0;
		
		for (File queryFile : queryFiles) {
			Document queryDoc = null;
			if (preloadedQueryDocs != null)
				queryDoc = preloadedQueryDocs.get(queryFile);
			if (queryDoc == null) {
				queryDoc = new Document(queryFile, config);
				queryDoc.createFingerprints(config, hashFreq, false);
			}
			
			Map<String, Integer> retrievedFilenamesNum = new HashMap<String, Integer>();
			List<Integer> hashes = queryDoc.getHashes();
			for (int i = 0; i < hashes.size() - config.k + 1; i++) {
				List<String> retrieved = searchDocs(searcher, hashes.subList(i, i + config.k), config);
				for (String filename : retrieved) {
					Integer num = retrievedFilenamesNum.get(filename);
					if (num == null)
						retrievedFilenamesNum.put(filename, 1);
					else
						retrievedFilenamesNum.put(filename, num + 1);
				}
			}
			
			// post-processing results
			Set<String> retrievedFilenames = new HashSet<String>();
			for (Entry<String, Integer> entry : retrievedFilenamesNum.entrySet())
				if (entry.getValue() >= 1) // currently, not used
					retrievedFilenames.add(entry.getKey());
			
			totalRetrievedFiles += retrievedFilenames.size();
			
			String queryFilename = queryFile.getName();
			Set<String> correctCollectionFilenames = correctPairs.get(queryFilename);
			
			Set<String> tpRetrieved = new HashSet<String>();
			
			for (String retrievedFilename : retrievedFilenames) {
				owPairsToAlign.write(queryFilename);
				owPairsToAlign.write(" ");
				owPairsToAlign.write(retrievedFilename);
				owPairsToAlign.write("\r\n");
				
				if ((correctCollectionFilenames != null) && correctCollectionFilenames.contains(retrievedFilename)) {
					tp++;
					tpRetrieved.add(retrievedFilename);
				} else {
					owPairsCorrect.write(queryFilename);
					owPairsCorrect.write(" ");
					owPairsCorrect.write(retrievedFilename);
					owPairsCorrect.write("\r\n");
					
					String xmlFilename = queryFilename.substring(0, queryFilename.length() - 4) + "-" + retrievedFilename.substring(0, retrievedFilename.length() - 4) + ".xml";
					OutputStreamWriter owXML = new OutputStreamWriter(new FileOutputStream(resultsPath + "correct/" + xmlFilename), StandardCharsets.UTF_8);
					owXML.write("<document reference=\"" + queryFilename + "\"></document>");
					owXML.close();
				}
			}
			
			if (correctCollectionFilenames == null)
				continue;
			
			for (String correctCollectionFilename : correctCollectionFilenames) {
				if (tpRetrieved.contains(correctCollectionFilename))
					continue;
				String xmlFilename = queryFilename.substring(0, queryFilename.length() - 4) + "-" + correctCollectionFilename.substring(0, correctCollectionFilename.length() - 4) + ".xml";
				OutputStreamWriter owXML = new OutputStreamWriter(new FileOutputStream(resultsPath + "out/" + xmlFilename), StandardCharsets.UTF_8);
				owXML.write("<document reference=\"" + queryFilename + "\"></document>");
				owXML.close();
			}
		}
		
		owPairsToAlign.close();
		owPairsCorrect.close();
		
		reader.close();
		indexDirectory.close();
		
		double denominator = totalRetrievedFiles;
		double precision = denominator > 0 ? tp / (double) denominator : 1;
		
		denominator = totalCorrectPairs;
		double recall = denominator > 0 ? tp / (double) denominator : 0;
		
		double f1 = (precision > 0) && (recall > 0) ? 2 * (precision * recall) / (precision + recall) : 0;
		
		return totalRetrievedFiles + "\t" + precision + "\t" + recall + "\t" + f1;
	}
	
	private List<String> searchDocs(IndexSearcher searcher, List<Integer> hashes, Configuration config) throws IOException {
		BooleanQuery.Builder builder = new BooleanQuery.Builder();
		
		for (int hash : hashes) {
			Query clause = new TermQuery(new Term(FIELD_FINGERPRINT, new BytesRef(Integer.toString(hash, 32).getBytes(StandardCharsets.US_ASCII))));
			builder.add(new BooleanClause(clause, BooleanClause.Occur.MUST));
		}
		
		Query query = builder.build();
		
		List<String> retrievedFilenames = new ArrayList<String>(config.nTop);
		
		TopDocs topDocs = searcher.search(query, config.nTop);
		ScoreDoc[] hits = topDocs.scoreDocs;
		for (ScoreDoc scoreDoc : hits) {
			DocumentStoredFieldVisitor visitor = new DocumentStoredFieldVisitor(FIELD_FILENAME);
			searcher.doc(scoreDoc.doc, visitor);
			org.apache.lucene.document.Document doc = visitor.getDocument();
			retrievedFilenames.add(doc.get(FIELD_FILENAME));
		}
		
		return retrievedFilenames;
	}
	
	private static List<File> getDirectoryFileList(String dirPath) throws IOException {
		File dir = new File(dirPath);
		List<File> files = new ArrayList<File>();
		if (dir.exists()) {
			File[] filesInDir = dir.listFiles();
			for (File file : filesInDir) {
				if (file.isDirectory())
					continue;
				files.add(file);
			}
		}
		return files;
	}
}
